package F5::ASMConfig::Entity::AllowedMethod;

use strict;
use warnings;

use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::ASMConfig;
use File::Basename;

use base qw/F5::ASMConfig::Entity::Base F5::ASMConfig::AllowedMethod/;

# NOTE: these constants should be equivalent to MCP enum mcp_cfg_src_t
use constant {
    USER_CFG => 0,
    BASE_CFG => 1,
};

use constant LOG_FILTER_KEY_HTTP_METHOD => 13; # from enum fw_log_filter_key_t

my @factory_methods;

#----------------------------------------------------------------------

my %STATIC_METHODS = ( GET => 'GET',
                       POST => 'POST',
                       HEAD => 'GET',);

sub before_store {
    my($self, %params) = @_;

    my $dbh = $self->{dbh};
    my $store_obj = $params{store_obj};
    my $store_mode = $params{store_mode};
    my $logger = $params{logger} || $self->{logger};
    my $call_flags = $params{call_flags} || {};

    if ($params{validate_value} && $self->has_value_changed(%params, check_fields => 'method_name')){
    	$self->validate_field(	field_value => $store_obj->{'method_name'},
				field_def => $self->fields->{'method_name'},
				field_name => 'method_name');
    }

    my $method_name = $self->get_new_value(%params,
                                           field_name => 'method_name', );

    if ($store_obj->{act_as_method} && $STATIC_METHODS{$method_name} && $store_obj->{act_as_method} ne $STATIC_METHODS{$method_name}) {
        die "Method $method_name can not be set to act as $store_obj->{act_as_method}\n";
    }

    if ($store_mode eq 'add' && ! $store_obj->{act_as_method} && 
        $store_obj->{method_name} && $store_obj->{method_name} =~ /\S/) {

        # Set by default_act_as
        my $method = $self->get_method(name => $store_obj->{method_name},
                                       add_if_missing => ($call_flags->{skip_mcp}) ? 0 : 1,
                                       call_flags => $call_flags,
                                       logger => $logger,
                                      );

        if ($method && $method->{default_act_as}) {
            $store_obj->{act_as_method}  = basename($method->{default_act_as});
        }

        $store_obj->{act_as_method} ||= 'GET';
    }

}

#----------------------------------------------------------------------

sub map_field_value {
    my($class, %params) = @_;
    my $field_name = $params{field_name};
    my $field_value = $params{field_value};
    my $field_def = $params{field_def};
    my $validate_value = defined $params{validate_value} ? $params{validate_value} : 1;
    my $call_flags = $params{call_flags} || {};

    if ($params{direction} eq 'to_db' && $validate_value) {
        if ($field_def->{normalize_func}) {
            my $rv = $class->validate_field(%params,
                                            field_value  => $field_value,
                                            validate_func => $field_def->{normalize_func},
                                            normalize_error_message => $field_def->{normalize_error_message},
                                           );
            $field_value = $rv->{normalized_value};
        }

        if ($field_def->{validate_func}) {
            $class->validate_field(%params,
                                   field_value  => $field_value,
                                  );
        }
    }   
    
    if ($params{direction} eq 'to_db' && ($field_name eq 'method_name' || $field_name eq 'act_as_method')) {
        my %additional_params;

        if ($field_name eq 'method_name' && $params{store_obj} && $params{store_obj}->{act_as_method}) {
            $additional_params{default_act_as} = $params{store_obj}->{act_as_method};
        }

        my $found_method = $class->get_method(name => $field_value,
                                              add_if_missing => ($call_flags->{skip_mcp}) ? 0 : 1,
                                              call_flags => $call_flags,
                                              logger => $params{logger},
                                              %additional_params,
                                             );
        if ($found_method) {
            return $found_method->{leaf_name};
        }
    }
    else {
        return $class->SUPER::map_field_value(%params);
    }
}

#----------------------------------------------------------------------


sub get_method {
  my($class, %params) = @_;
  $params{leaf_name} ||= $params{name};
  my $method_name = $params{leaf_name};
  #method_code left here for upgrade/import usage
  my $method_code = $params{code};
  my $search_field = (defined $method_code ? 'code' : 'leaf_name');

  $class->initialize_cache_if_needed(%params);

  my $found_method;
  if ($params{$search_field}) {
      ($found_method) = grep { $_->{$search_field} eq $params{$search_field} } @factory_methods;

      if (! $found_method) {
          require F5::ASM::COAPI;
          my $found_methods = F5::ASM::COAPI::Query->new('asm_http_method', { $search_field => $params{$search_field} });
          $found_method = $found_methods->[0] if ($found_methods && @$found_methods);
      }
  }

  if (! $found_method && $method_name && $params{add_if_missing}) {
      require F5::ASM::COAPI;
      my $new_method = F5::ASM::COAPI::Object->new("asm_http_method");
      $new_method->{name} = $method_name;
      $new_method->{default_act_as} = $params{default_act_as} || 'GET';
      $new_method->save();
      $params{call_flags}->{writes_to_mcp} = 1 if defined $params{call_flags};

      # Persist configuration (before restart mcpd)
      F5::ASM::COAPI::SaveAll(logger => $params{logger});

      # Retrieve auto-incremented code
      my $found_methods = F5::ASM::COAPI::Query->new('asm_http_method', { leaf_name => $method_name });
      $found_method = $found_methods->[0] if ($found_methods && @$found_methods);
  }

  return $found_method;
}

#----------------------------------------------------------------------

my @required_methods = qw(GET POST);

sub initialize_cache_if_needed {
    my($class, %params) = @_;

    return if (@factory_methods);
    require F5::ASM::COAPI;
    @factory_methods = @{F5::ASM::COAPI::Query->new('asm_http_method', {config_source => BASE_CFG})};
    #Sanity check that at least GET and POST exist in MCP (required to add any others)
    for my $method (@required_methods) {
        if (! grep { $_->{leaf_name} eq $method } @factory_methods) {
            my $new_method = F5::ASM::COAPI::Object->new("asm_http_method");
            $new_method->{name} = $method;
            $new_method->{default_act_as} = $method;
            $new_method->save();
            push (@factory_methods, @{F5::ASM::COAPI::Query->new('asm_http_method', {leaf_name => $method})});
        }
    }
}

#----------------------------------------------------------------------

sub get_method_name {
    my($class_or_self, %params) = @_;

    if (ref($class_or_self) && $class_or_self->{dbh} && ! $params{dbh}) {
        $params{dbh} = $class_or_self->{dbh};
    }

    my $found_method = $class_or_self->get_method(%params);
    if ($found_method) {
        return $found_method->{leaf_name};
    }
    return;
}

#----------------------------------------------------------------------

my %REQUIRED_METHODS = ( GET => 1,
                         POST => 1,);

sub is_deletable {
    my($self, %params) = @_;
    if ($REQUIRED_METHODS{$self->{method_name}}) {
        my $msg = "You can not delete required (GET or POST) method.";
        return { error_message => $msg,
                 display_message => $msg,
                 error_level     => 'info',
                 error_code      => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
               };
    }
    return $self->SUPER::is_deletable(%params);
}

#----------------------------------------------------------------------

sub delete {
    my($self, %params) = @_;

    my $rv = $self->SUPER::delete(%params);
    my $call_flags = $params{call_flags} || {};
    $self->cleanup_user_defined_methods(dbh => $self->{dbh}) unless $call_flags->{skip_mcp};

    return $rv;
}

#----------------------------------------------------------------------

sub cleanup_user_defined_methods {
    my($class_or_self, %params) = @_;

    if (ref($class_or_self) && $class_or_self->{dbh} && ! $params{dbh}) {
        $params{dbh} = $class_or_self->{dbh};
    }

    require F5::ASM::COAPI;
    my $dbh = $params{dbh};
    my $allowed_policy_methods = $dbh->selectcol_arrayref("SELECT DISTINCT method_name FROM PLC.PL_ALLOWED_METHODS");
    my $allowed_url_methods = $dbh->selectcol_arrayref("SELECT DISTINCT method_name FROM PLC.PL_OBJECT_METHOD_OVERRIDES");
    my $http_method_filters = F5::ASM::COAPI::Query->new('fw_log_request_filter', {key => LOG_FILTER_KEY_HTTP_METHOD});
    my $httpsecurity_profiles = F5::ASM::COAPI::Query->new('profile_httpsecurity');

    my $user_defined_methods = F5::ASM::COAPI::Query->new('asm_http_method', {config_source => USER_CFG});

    for my $user_defined_method (@$user_defined_methods) {
        next if (grep { $_ eq $user_defined_method->{leaf_name} } @$allowed_policy_methods);
        next if (grep { grep { $_ eq $user_defined_method->{leaf_name} } @{$_->{value_list}} } @$http_method_filters);
        next if (grep { grep { $_ eq $user_defined_method->{leaf_name} } @{$_->{methods}} } @$httpsecurity_profiles);
        next if (grep { $_ eq $user_defined_method->{leaf_name} } @$allowed_url_methods);
        $user_defined_method->delete();
    }
}

#----------------------------------------------------------------------

# objects, which are not fully in the mcp_object model, override the mcp_folder function
sub mcp_folder {
    return '/Common';
}

#----------------------------------------------------------------------

1;
